/*jslint nomen: true, maxlen: 200, indent: 2, maxerr: 100*/
/*jshint esversion: 6*/
/*global window, document, require, URL, rJS*/

(function (window, document, rJS) {
  "use strict";
  let repo = "nexedi/slapos";
  let branch = "master";
  const defaultSchema = `https://lab.nexedi.com/${repo}/raw/${branch}/software/kvm/instance-kvm-input-schema.json`;

  let monacoSchemaEditor;
  let monacoEditor;
// monaco should ignore edit event when it's caused by rjs_json_form
  let monacoIgnoreEvent = false;
  /*
   based on https://microsoft.github.io/monaco-editor/playground.html#extending-language-services-configure-json-defaults
   and https://codepen.io/maxvrko/full/XVxVyY
  */

  require.config({
    paths: {vs: "https://unpkg.com/monaco-editor@0.15.5/min/vs"}
  });
  let proxy = URL.createObjectURL(
    new Blob(
      [`
self.MonacoEnvironment = {
  baseUrl: 'https://unpkg.com/monaco-editor@0.15.5/min/'
};
importScripts('https://unpkg.com/monaco-editor@0.15.5/min/vs/base/worker/workerMain.js');
`
      ],
      {type: "text/javascript"}
    )
  );

  window.MonacoEnvironment = {getWorkerUrl: () => proxy};

  const getSchema = (schemaUrl, mappedUrl)=> {
    return fetch(schemaUrl)
      .then(resp => {
        return resp.json();
      })
      .then(s => {
        return {
          uri: mappedUrl || schemaUrl,
          schema: s
        };
      });
  };

  rJS(window)
    .allowPublicAcquisition("notifyValid", function (arr, scope) {
    })
    .allowPublicAcquisition("notifyInvalid", function (arr, scope) {
    })
    .allowPublicAcquisition("notifyChange", function (arr, scope) {
      return this.getDeclaredGadget("rjs_json_form").then(function (g) {
        return g.getContent().then(function (content) {
          monacoIgnoreEvent = true;
          monacoEditor
            .getModel()
            .setValue(JSON.stringify(JSON.parse(content.rjs_json_form), null, 4));
          monacoIgnoreEvent = false;
        });
      });
    })
    .ready(function () {
      let g = this;
      g.props = {};
      document.getElementById("monaco-schema-url").value = defaultSchema;

      require(["vs/editor/editor.main"], function (monaco) {
        g.props.monaco = monaco;
        g.props.modelUri = monaco.Uri.parse("tmp://xxx.json"); // a made up unique URI for our model
        let model = monaco.editor.createModel("{\n}", "json", g.props.modelUri);

        Promise.all([
          getSchema("../../json-schema/schema7.json", "http://json-schema.org/draft-07/schema")
        ]).then(schemas => {
          g.props.baseSchemas = schemas;

          monacoEditor = monaco.editor.create(
            document.getElementById("monaco-editor"),
            {
              model: model
            }
          );

          monacoEditor.onDidChangeModelContent(e => {
            if (!monacoIgnoreEvent) {
              console.log("monaco changed", monacoEditor.getValue());
              let parsed;
              try {
                parsed = JSON.parse(monacoEditor.getValue());
              } catch (error) {
                console.error("parse error, ignoring changes from monaco", error);
                return;
              }
              g.getDeclaredGadget("rjs_json_form").then(g => {
                return g.render({
                  value: parsed
                });
              });
            }
          });

          g.changeState({url: defaultSchema});
        });
      });
    })
    .onStateChange(function () {
      let g = this;
      return getSchema(g.state.url).then(s => {
        s.fileMatch = [g.props.modelUri.toString()]; // this is the main schema associated to the model
        return s;
      }).then(schema => {
        g.props.monaco.languages.json.jsonDefaults.setDiagnosticsOptions({
          validate: true,
          enableSchemaRequest: true,
          schemas: [schema].concat(g.props.baseSchemas)
        });
        monacoIgnoreEvent = true;
        if (monacoSchemaEditor) {
          monacoSchemaEditor
            .getModel()
            .setValue(JSON.stringify(schema.schema, null, 4));
        } else {
          monacoSchemaEditor = g.props.monaco.editor.create(
            document.getElementById("monaco-schema"),
            {
              model: g.props.monaco.editor.createModel(JSON.stringify(schema.schema, null, 4), "json", schema.uri)
            }
          );

          monacoSchemaEditor.onDidChangeModelContent(e => {
            if (!monacoIgnoreEvent) {
              console.log("schema changed", monacoEditor.getValue());
              let parsed;
              try {
                parsed = JSON.parse(monacoSchemaEditor.getValue());
              } catch (error) {
                console.error("parse error, ignoring changes from monaco", error);
                return;
              }
              g.getDeclaredGadget("rjs_json_form").then(g => {
                return g.render({
                  schema: parsed
                });
              });
            }
          });
        }
        monacoIgnoreEvent = false;
        return g.getDeclaredGadget("rjs_json_form").then(gadget => {
          return gadget.render({
            key: "rjs_json_form",
            schema: schema.schema,
            schema_url: g.state.url
          });
        });
      })
        .catch(e => {

        });
    })
    .onEvent('input', function (evt) {
      if (evt.target.id === "monaco-schema-url") {
        return this.changeState({
          url: evt.target.value
        });
      }
    });
}(window, document, rJS));